package cn.jlsxwkj.moudles.chat;

import cn.dev33.satoken.SaManager;
import cn.hutool.core.lang.Tuple;
import cn.hutool.crypto.digest.MD5;
import cn.jlsxwkj.common.config.UserConfig;
import cn.jlsxwkj.common.exception.*;
import cn.jlsxwkj.common.utils.FileType;
import cn.jlsxwkj.common.utils.Log;
import cn.jlsxwkj.common.utils.MergeDocuments;
import cn.jlsxwkj.moudles.chat.message.*;
import cn.jlsxwkj.moudles.chathistory.ChatHistoryService;
import jakarta.annotation.Resource;
import org.springframework.ai.document.Document;
import org.springframework.ai.document.DocumentReader;
import org.springframework.ai.embedding.EmbeddingModel;
import org.springframework.ai.vectorstore.SearchRequest;
import org.springframework.ai.vectorstore.VectorStore;
import org.springframework.stereotype.Service;
import org.springframework.web.multipart.MultipartFile;
import reactor.core.publisher.Flux;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;
import java.util.Objects;
import java.util.stream.Collectors;


/**
 * @author zh
 */
@Service
public class ChatService {

    private final MD5 md5 = cn.hutool.crypto.digest.MD5.create();
    private final String PATH = System.getProperty("user.dir") + "/ai_doc/";
    private List<Message> listMessage;
    private StringBuffer chatMessage;

    @Resource
    private VectorStore vectorStore;
    @Resource
    private EmbeddingModel embeddingModel;
    @Resource
    private ChatUtil chatUtil;
    @Resource
    private ChatHistoryService chatHistoryService;
    @Resource
    private UserConfig userConfig;


    /**
     * 使用spring ai解析txt文档并保存至 pg
     *
     * @param file 保存的文件
     * @return 保存状态
     */
    public String uploadDocument(MultipartFile file) throws CustomException {
        Tuple supportFile = isSupportFile(file);
        File saveFile = supportFile.get(0);
        File fileName = supportFile.get(1);
        DocumentReader reader = supportFile.get(3);
        if (!saveFile.exists()) {
            try {
                file.transferTo(saveFile);
                vectorStore.add(reader.get());
                return "上传完成";
            } catch (Exception e) {
                boolean delete = saveFile.delete();
                if (!delete) {
                    Log.warn(this.getClass(), "删除文件失败: {}", fileName);
                    throw new UnknownException();
                }
            }
        }
        return "上传完成";
    }

    /**
     * 根据关键词搜索向量库
     *
     * @param keyword 关键词
     * @return 文本内容
     */
    public String search(String keyword) {
        SearchRequest searchRequest = SearchRequest.query(keyword).withSimilarityThreshold(0.5);
        List<Document> documents = vectorStore.similaritySearch(searchRequest);
        return mergeDocuments(documents);
    }

    /**
     * 问答流,根据输入内容回答
     *
     * @param message 输入内容
     * @return 回答内容
     */
    public Flux<String> chatStream(String message) {
        String context = search(message);
        return stream(message, new SystemMessage(context));
    }

    /**
     * 问答总结流, 根据给定的文档总结内容
     *
     * @param message 输入内容
     * @param file    文件
     * @return 内容总结
     */
    public Flux<String> chatSummaryStream(String message, MultipartFile file) {
        embeddingModel.embed("");
        Tuple supportFile;
        try {
            supportFile = isSupportFile(file);
        } catch (CustomException e) {
            return Flux.just(e.getMessage()).concatWithValues("<{完成}>");
        }
        String fileName = supportFile.get(1);
        String fileTypeCn = supportFile.get(2);
        DocumentReader reader = supportFile.get(3);
        String format = """
                %s :: %s ====> %s
                ↑ %s
                """;
        String context = String.format(format, fileTypeCn, fileName, mergeDocuments(reader.get()), message);
        UserMessage userMessage = new UserMessage(context);
        Log.info(this.getClass(), context);
        return stream(message, userMessage);
    }

    /**
     * 合并文档
     *
     * @param documents 文档列表
     * @return 合并后的文档
     */
    private String mergeDocuments(List<Document> documents) {
        return MergeDocuments.mergeDocuments(documents).stream()
                .map(Document::getContent)
                .collect(Collectors.joining("\n"));
    }

    /**
     * 停止 chat
     */
    public void stop() {
        chatUtil.stop();
    }

    /**
     * 获取prompt
     *
     * @param message 提问内容
     * @param context 提示词
     */
    private void setChatMessage(String message, Message context) {
        if (listMessage == null) {
            listMessage = new ArrayList<>();
            listMessage.add(new SystemMessage(userConfig.getSysMessage()));
        }
        if (listMessage.size() > userConfig.getMaxListMessageLength()) {
            int start = listMessage.size() - userConfig.getMaxListMessageLength();
            listMessage = listMessage.subList(start, listMessage.size()).stream()
                    .filter(msg -> !msg.getContent().equals(userConfig.getSysMessage()))
                    .map(msg -> {
                        if (msg.getRole().equals(UserRole.ASSISTANT)) {
                            String content = msg.getContent();
                            if (content.length() > userConfig.getMaxMessageLength()) {
                                String assistantMessage = content.substring(0, userConfig.getMaxMessageLength());
                                return new AssistantMessage(assistantMessage);
                            }
                        }
                        return msg;
                    }).distinct()
                    .collect(Collectors.toList());
            listMessage.add(new SystemMessage(userConfig.getSysMessage()));
        }
        if (context.getRole().equals(UserRole.SYSTEM)) {
            listMessage.add(new UserMessage(message));
            if (!context.getContent().isBlank()) {
                listMessage.add(context);
            }
        }
        if (context.getRole().equals(UserRole.USER)) {
            listMessage.add(context);
        }
    }

    /**
     * chat stream 封装
     *
     * @param message 提问内容
     * @param context 提示词
     * @return flux
     */
    private Flux<String> stream(String message, Message context) {
        String userId = SaManager.getStpLogic("").getLoginIdAsString();
        chatMessage = new StringBuffer();
        setChatMessage(message, context);
        return chatUtil.chat(listMessage, true).map(response -> {
            String str = response.getMessage().getContent();
            chatMessage.append(str);
            System.out.print(str);
            return str.replace(" ", "  ")
                    .replace("\t", "    ");
        }).concatWithValues("<{完成}>")
                .cancelOn(chatUtil.getScheduler())
                .doOnCancel(this::stop)
                .doOnComplete(() -> {
                    String chatMsg = chatMessage.toString();
                    listMessage.add(new AssistantMessage(chatMsg));
                    try {
                        chatHistoryService.insert(userId, message, chatMsg);
                    } catch (InsertFailException e) {
                        Log.warn(this.getClass(), e.getMessage());
                    }
                    Log.info(this.getClass(), "{} user message ==> {}", userId, message);
                    Log.info(this.getClass(), "{} chat message ==> {}", userId, chatMsg);
                });
    }

    /**
     * 封装文件
     *
     * @param file 文件流
     * @return 元组
     */
    private Tuple isSupportFile(MultipartFile file) throws CustomException {
        byte[] bytes;
        try {
            bytes = file.getBytes();
        } catch (IOException e) {
            throw new OpenFileFailException("读取文件失败");
        }
        String[] split = Objects.requireNonNull(file.getOriginalFilename()).split("\\.");
        String fileType = split[split.length - 1].toLowerCase(Locale.ROOT);
        File savePath = new File(PATH);
        File saveFile = new File(PATH + md5.digestHex(bytes) + "." + fileType);
        String fileName = saveFile.getName();
        if (!savePath.exists()) {
            if (!savePath.mkdirs()) {
                Log.warn(this.getClass(), "创建文件夹失败: " + PATH);
            }
        }
        Tuple fileTypeAndResource = new FileType(bytes, fileName, 5, userConfig.getCnocrUrl())
                .getFileTypeAndResource(fileType);
        if (fileTypeAndResource == null) {
            throw new FileTypeDoesNotSupportException("暂不支持的文件类型: " + fileType);
        }
        String fileTypeCn = fileTypeAndResource.get(0);
        DocumentReader reader = fileTypeAndResource.get(1);
        return new Tuple(saveFile, fileName, fileTypeCn, reader);
    }
}
